library in_networkclient;

(* ***** BEGIN LICENSE BLOCK *****
 * Version: MPL 1.1
 *
 * The contents of this file are subject to the Mozilla Public License Version 
 * 1.1 (the "License"); you may not use this file except in compliance with 
 * the License. You may obtain a copy of the License at 
 * http://www.mozilla.org/MPL/
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 * for the specific language governing rights and limitations under the
 * License.
 *
 * The Original Code is the Network Client Input Module for Lightning.
 *
 * The Initial Developer of the Original Code is
 * Tomsoft.
 * Portions created by Tomsoft are Copyright (C) 2012
 * Tomsoft. All Rights Reserved.
 *
 * Contributor(s):
 *   Tommy Sools <tom@sools.com>
 *
 * ***** END LICENSE BLOCK ***** *)

{$IMAGEBASE $410E0000}

uses
  SysUtils,
  Classes,
  Windows,
  Math,
  StdCtrls,
  Graphics,
  WinSock,
  common in '..\common.pas',
  WSockets in '..\components\WSockets.pas';

{$R *.RES}

const
  channels = 32;
  defaultPort = 80055;
  maxQueue = 10000;

type
  TPacket = record
    channel: 1..channels;
    level:   TLevel;
  end;
  TDetails = class(TObject)
  public
    gbDetails:     TGroupBox;
    lHost:         TLabel;
    eHost:         TEdit;
    lPort:         TLabel;
    ePort:         TEdit;
    bUpdate:       TButton;
    stConnected:   TStaticText;
    chBinary:      TCheckBox;
    lReconnect:    TLabel;
    eReconnect:    TEdit;
    bTCPClient:    TTCPClient;
    levels:        array[1..channels] of TLevel;
    status:        array[1..channels] of Boolean;
    queue:         array of TPacket;
    binaryQueue:   array of Integer;
    binary:        Boolean;
    reconnect:     Integer;
    lastReconnect: Int64;
    procedure bUpdateClick(Sender: TObject);
    procedure chBinaryClick(Sender: TObject);
    procedure eReconnectChange(Sender: TObject);
    procedure tcpClientConnect(Sender: TObject; Socket: TSocket);
    procedure tcpClientClose(Sender: TObject; Socket: TSocket);
    procedure tcpClientData(Sender: TObject; Socket: TSocket);
  end;

procedure TDetails.bUpdateClick(Sender: TObject);
begin
  bTCPClient.Host := eHost.Text;
  bTCPClient.Port := ePort.Text;
  if bTCPClient.SocketState <> ssClosed then
    bTCPClient.Close();
  if Trim(bTCPClient.Host) <> '' then
    bTCPClient.Open();
end;

procedure TDetails.chBinaryClick(Sender: TObject);
begin
  binary := chBinary.Checked;
  if bTCPClient.SocketState <> ssClosed then
    bTCPClient.Close();
  if Trim(bTCPClient.Host) <> '' then
    bTCPClient.Open();
end;

procedure TDetails.eReconnectChange(Sender: TObject);
begin
  reconnect := StrToInt(eReconnect.Text);
end;

procedure TDetails.tcpClientConnect(Sender: TObject; Socket: TSocket);
begin
  if gbDetails <> nil then
  begin
    stConnected.Font.Color := clGreen;
    stConnected.Caption    := 'Connected.';
  end;
end;

procedure TDetails.tcpClientClose(Sender: TObject; Socket: TSocket);
begin
  if gbDetails <> nil then
  begin
    stConnected.Font.Color := clRed;
    stConnected.Caption    := 'Not connected.';
  end;
end;

procedure TDetails.tcpClientData(Sender: TObject; Socket: TSocket);
var
  packet: TPacket;
  i, r, int: Integer;
begin
  if binary then
  begin
    repeat
      r := bTCPClient.ReadBuffer(@int, SizeOf(int));
      if r = SizeOf(int) then
      begin
        i := Length(binaryQueue);
        SetLength(binaryQueue, i + 1);
        binaryQueue[i] := int;
      end;
    until r <> SizeOf(packet);
    if Length(binaryQueue) > maxQueue then
    begin
      Windows.Beep(400, 100);
      SetLength(binaryQueue, 0);
    end;
  end
  else
  begin
    repeat
      r := bTCPClient.ReadBuffer(@packet, SizeOf(packet));
      if r = SizeOf(packet) then
      begin
        i := Length(queue);
        SetLength(queue, i + 1);
        queue[i] := packet;
      end;
    until r <> SizeOf(packet);
    if Length(queue) > maxQueue then
    begin
      Windows.Beep(400, 100);
      SetLength(queue, 0);
    end;
  end;
end;

function  init(name: PChar; nameLength: DWORD): DWORD; cdecl;
begin
  Result := return('INetworkClient', name, nameLength);
  tickInit;
end;

procedure finish(); cdecl;
begin
  //
end;

function  create(controlHandle: THandle): PLightningModule; cdecl;
var
  myDetails: TDetails;
  i:         Integer;
begin
  New(Result);
  with Result^ do
  begin
    handle        := controlHandle;
    title         := 'NetworkClient';
    settings      := True;
    width         := -1;
    height        := -1;
    inputTop      := -1;
    infoTop       := -1;
    outputTop     := -1;
    inputs        := 0;
    outputs       := channels;
    inputTooltip  := '';
    GetMem(outputTooltip, maxLength);
    StrCopy(outputTooltip, '');
    for i := 1 to channels do
      StrCat(outputTooltip, PChar('Channel ' + IntToStr(i) + ' value|'));
    myDetails     := TDetails.Create();
    reserved      := nil;
    details       := myDetails;
  end;
  with myDetails do
  begin
    binary        := False;
    reconnect     := 0;
    lastReconnect := 0;
    for i := 0 to channels - 1 do
    begin
      levels[i] := levelMin;
      status[i] := False;
    end;
    gbDetails := nil;
    bTCPClient           := TTCPClient.Create(nil);
    bTCPClient.OnConnect := tcpClientConnect;
    bTCPClient.OnClose   := tcpClientClose;
    bTCPClient.OnData    := tcpClientData;
    bTCPClient.Port      := IntToStr(defaultPort);
  end;
end;

procedure destroy(lightningModule: PLightningModule); cdecl;
begin
  FreeMem(lightningModule^.outputTooltip);
  TDetails(lightningModule^.details).bTCPClient.Free;
  Dispose(lightningModule);
end;

function  action(lightningModule: PLightningModule; input, output: Integer; inputConnected, outputConnected: PChar; var bridge: TSpectrum): Boolean; cdecl;
var
  details:  TDetails;
  i, check: Integer;
  tick:     Int64;
begin
  Result  := True;
  details := TDetails(lightningModule^.details);
 if details.bTCPClient.SocketState = ssConnected then
  if details.binary then
  begin
    if output = 1 then // if this is the first output
    begin
      if Length(details.binaryQueue) > 0 then
      begin
        check := 0;
        for i := 1 to channels do
        begin
          if i = 1 then
            check := 1
          else
            check := check * 2;
          details.status[i] := (details.binaryQueue[0] and check) = check;
        end;
        for i := 1 to Length(details.binaryQueue) - 1 do
          details.binaryQueue[i - 1] := details.binaryQueue[i];
        SetLength(details.binaryQueue, Length(details.binaryQueue) - 1);
      end;
    end;
    if details.status[output] then
      createSpectrum(bridge, levelMax)
    else
      createSpectrum(bridge, levelMin);
  end
  else
  begin
    createSpectrum(bridge, details.levels[output]);
    if Length(details.queue) > 0 then
      if details.queue[0].channel = output then
      begin
        details.levels[output] := details.queue[0].level;
        createSpectrum(bridge, details.queue[0].level);
        for i := 1 to Length(details.queue) - 1 do
          details.queue[i - 1] := details.queue[i];
        SetLength(details.queue, Length(details.queue) - 1);
      end;
  end
 else
 begin
   createSpectrum(bridge, levelMin);
   if details.reconnect > 0 then
   begin
     tick := tickCount;
     if tick > details.lastReconnect + details.reconnect * 1000 then
     begin
       details.lastReconnect := tick;
       if details.bTCPClient.SocketState <> ssClosed then
         details.bTCPClient.Close();
       if Trim(details.bTCPClient.Host) <> '' then
         details.bTCPClient.Open();
     end;
   end;
 end;
end;

procedure show(lightningModule: PLightningModule; handle: THandle); cdecl;
begin
  with TDetails(lightningModule^.details) do
  begin
    gbDetails := TGroupBox.Create(nil);
    with gbDetails do
    begin
      Top     := detailsTop;
      Left    := detailsLeft;
      Width   := detailsWidth;
      Height  := detailsHeight;
      Ctl3D   := True;
      Color   := clBtnFace;
      Caption := 'Network Client Settings';
    end;
    lHost := TLabel.Create(gbDetails);
    with lHost do
    begin
      AutoSize := True;
      Top      := 20;
      Left     := 8;
      Caption  := 'Server host:';
      Parent   := gbDetails;
    end;
    eHost := TEdit.Create(gbDetails);
    with eHost do
    begin
      Top    := 40;
      Left   := 8;
      Width  := 120;
      Text   := bTCPClient.Host;
      Parent := gbDetails;
    end;
    lPort := TLabel.Create(gbDetails);
    with lPort do
    begin
      AutoSize := True;
      Top      := 70;
      Left     := 8;
      Caption  := 'Server port:';
      Parent   := gbDetails;
    end;
    ePort := TEdit.Create(gbDetails);
    with ePort do
    begin
      Top    := 90;
      Left   := 8;
      Width  := 120;
      Text   := bTCPClient.Port;
      Parent := gbDetails;
    end;
    bUpdate := TButton.Create(gbDetails);
    with bUpdate do
    begin
      Top     := 90;
      Left    := 140;
      Width   := 120;
      Height  := eHost.Height;
      Caption := 'Update';
      OnClick := bUpdateClick;
      Parent  := gbDetails;
    end;
    stConnected := TStaticText.Create(gbDetails);
    with stConnected do
    begin
      AutoSize   := True;
      Top        := 43;
      Left       := 140;
      Font.Style := [fsBold];
      if bTCPClient.SocketState <> ssConnected then
      begin
        stConnected.Font.Color := clRed;
        stConnected.Caption    := 'Not connected.';
      end
      else
      begin
        stConnected.Font.Color := clGreen;
        stConnected.Caption    := 'Connected.';
      end;
      Parent     := gbDetails;
    end;
    chBinary := TCheckBox.Create(gbDetails);
    with chBinary do
    begin
      Top     := 120;
      Left    := 8;
      Width   := 300;
      Caption := 'Binary mode (for better performance)';
      Checked := binary;
      OnClick := chBinaryClick;
      Parent  := gbDetails;
    end;
    lReconnect := TLabel.Create(gbDetails);
    with lReconnect do
    begin
      AutoSize := True;
      Top      := 20;
      Left     := 272;
      Caption  := 'If disconnected, reconnect every (seconds):';
      Parent   := gbDetails;
    end;
    eReconnect := TEdit.Create(gbDetails);
    with eReconnect do
    begin
      Top      := 40;
      Left     := 272;
      Width    := 50;
      Text     := IntToStr(reconnect);
      OnChange := eReconnectChange;
      Parent   := gbDetails;
    end;
    gbDetails.ParentWindow := handle;
  end;
end;

procedure hide(lightningModule: PLightningModule); cdecl;
begin
  with TDetails(lightningModule^.details) do
  begin
    eReconnect.Free();
    lReconnect.Free();
    chBinary.Free();
    stConnected.Free();
    bUpdate.Free();
    ePort.Free();
    lPort.Free();
    eHost.Free();
    lHost.Free();
    gbDetails.Free();
    gbDetails := nil;
  end;
end;

procedure open(lightningModule: PLightningModule; settings: PChar); cdecl;
var
  bString, bString2: string;
begin
  bString := settings;
  with TDetails(lightningModule^.details) do
  begin
    bTCPClient.Host := parse(bString, ',');
    parse(bString, ',', bString2);
    if Trim(bString2) = '' then
      bString2 := IntToStr(defaultPort);
    bTCPClient.Port := bString2;
    binary    := parse(bString, ',') <> '';
    reconnect := StrToIntDef(parse(bString, ','), 0);
    if bTCPClient.SocketState <> ssClosed then
      bTCPClient.Close();
    if Trim(bTCPClient.Host) <> '' then
      bTCPClient.Open();
  end;
end;

function  save(lightningModule: PLightningModule; settings: PChar; settingsLength: DWORD): DWORD; cdecl;
var
  bString: string;
begin
  with TDetails(lightningModule^.details) do
  begin
    bString := bTCPClient.Host + ',' + bTCPClient.Port + ',';
    if binary then
      bString := bString + '1';
    Result := return(bString + ',' + IntToStr(reconnect), settings, settingsLength);
  end;
end;

exports
  init, finish, create, destroy, action, show, hide, open, save;

begin
end.
